// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.cardboard;

import java.util.EventObject;
import java.util.List;
import java.io.Serializable;

/**
 * An event for a CardBoard.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastModifiedBy$
 * @version $Id: CardBoardEvent.java 129 2004-04-15 05:00:43Z jerome $
 */
public class CardBoardEvent extends EventObject
{
  /**
   * Identifies one or more changes in the lists contents.
   */
  public static final int CONTENTS_CHANGED = 0;
  /**
   * Identifies the addition of one or more contiguous items to the list.
   */
  public static final int INTERVAL_ADDED = 1;
  /**
   * Identifies the removal of one or more contiguous items from the list.
   */
  public static final int INTERVAL_REMOVED = 2;

  private final int _type;
  private final BoardIndexes _boardIndexes;

  /**
   * A supertype tag for the different indexes implementation.
   * @see CardBoardEvent.IndexInterval
   * @see CardBoardEvent.IndexList
   */
  public static interface BoardIndexes extends Serializable
  {
  }

  /**
   * A class that allows to specify BoardIndexes as an interval between two indexes.
   */
  public static final class IndexInterval implements BoardIndexes
  {
    private final int _index0;
    private final int _index1;

    /**
     * Creates an instance of IndexInterval given the two specified index boundaries.
     * @param index0
     * @param index1
     * @throws IllegalArgumentException if <code>index0 > index1</code>
     */
    private IndexInterval(final int index0, final int index1)
    {
      if (index0 > index1)
      {
        throw new IllegalArgumentException("_index0=" + index0 + " > _index1=" + index1);
      }
      this._index0 = index0;
      this._index1 = index1;
    }

    /**
     * @return the lower bound of the index interval
     */
    public int getIndex0()
    {
      return _index0;
    }

    /**
     * @return the higher bound of the index interval
     */
    public int getIndex1()
    {
      return _index1;
    }

    /**
     * Returns a string representation of this IndexInterval. This method
     * is intended to be used only for debugging purposes, and the
     * content and format of the returned string may vary between
     * implementations. The returned string may be empty but may not
     * be <code>null</code>.
     *
     * @return a string representation of this IndexInterval.
     */
    public String toString()
    {
      return _index0 + "-" + _index1;
    }
  }

  /**
   * A class that allows to specify BoardIndexes as a list of indexes.
   */
  public static final class IndexList implements BoardIndexes
  {
    private final List _indexesInBoard;

    /**
     * Creates an instance given the specified list of indexes.
     * @param indexesInBoard a <code>List</code> of <code>Integer</code>s.
     */
    private IndexList(final List indexesInBoard)
    {
      _indexesInBoard = indexesInBoard;
    }

    /**
     * Returns the list of indexes on the board.
     * @return a <code>List</code> of <code>Integer</code>s.
     */
    public List getIndexesInBoard()
    {
      return _indexesInBoard;
    }
    /**
     * Returns a string representation of this IndexInterval. This method
     * is intended to be used only for debugging purposes, and the
     * content and format of the returned string may vary between
     * implementations. The returned string may be empty but may not
     * be <code>null</code>.
     *
     * @return a string representation of this IndexInterval.
     */
    public String toString()
    {
      final StringBuffer buf = new StringBuffer();
      for (int i = 0; i < _indexesInBoard.size(); i++)
      {
        final Integer integer = (Integer) _indexesInBoard.get(i);
        if (i > 0)
        {
          buf.append(",");
        }
        buf.append(integer.intValue());
      }
      return buf.toString();
    }
  }

  /**
   * Returns the event _type. The possible values are:
   * <ul>
   * <li> {@link #CONTENTS_CHANGED}
   * <li> {@link #INTERVAL_ADDED}
   * <li> {@link #INTERVAL_REMOVED}
   * </ul>
   *
   * @return an int representing the _type value
   */
  public int getType()
  {
    return _type;
  }


  /**
   * Return the indexes of the cards modified on the board.
   * Can be an instance of {@link IndexInterval range} or a {@link
   * IndexList list of indexes}.
   * @return the board indexes
   */
  public BoardIndexes getBoardIndexes()
  {
    return _boardIndexes;
  }

  /**
   * Constructs a CardBoardEvent object.
   *
   * @param source the source Object (typically <code>this</code>)
   * @param type   an int specifying {@link #INTERVAL_ADDED}, or {@link #INTERVAL_REMOVED}
   * @param index0 an int specifying the bottom of a range
   * @param index1 an int specifying the top of a range
   * @throws IllegalArgumentException if {@link #CONTENTS_CHANGED} is specified as _type.
   */
  public CardBoardEvent(final Object source, final int type, final int index0, final int index1)
  {
    super(source);
    if (type == CONTENTS_CHANGED)
    {
      final String s = "CONTENTS_CHANGED type not allowed when specifying an interval.";
      throw new IllegalArgumentException(s);
    }
    this._type = type;
    this._boardIndexes = new IndexInterval(index0, index1);
  }

  /**
   * Constructs a CardBoardEvent object.
   *
   * @param source the source Object (typically <code>this</code>)
   * @param indexes a List of Integer int specifying the indexes changed.
   */
  public CardBoardEvent(final Object source, final List indexes)
  {
    super(source);
    this._type = CONTENTS_CHANGED;
    // FIXME we should copy that list and make it immutable
    this._boardIndexes = new IndexList(indexes);
  }

  /**
   * Returns a string representation of this CardBoardEvent. This method
   * is intended to be used only for debugging purposes, and the
   * content and format of the returned string may vary between
   * implementations. The returned string may be empty but may not
   * be <code>null</code>.
   *
   * @return a string representation of this CardBoardEvent.
   * @since 1.4
   */
  public String toString()
  {
    return getClass().getName()
         + "[_type=" + _type
         + ",{indexes=" + _boardIndexes + "}]";
  }
}
