// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.List;

/**
 * An implementation of a set game board.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: BoardImpl.java 129 2004-04-15 05:00:43Z jerome $
 */
public class BoardImpl implements SetGameBoard
{
  private final Logger _logger = Logger.getLogger(BoardImpl.class);
  /** the listeners. */
  private final List _boardListeners = new ArrayList();
  /** the cards. */
  private final List _board = new java.util.Vector(12);

  /**
   * @return the cards on the board.
   */
  public synchronized CardProperties[] getCards()
  {
    CardProperties[] props = new CardProperties[_board.size()];
    props = (CardProperties[]) _board.toArray(props);
    return props;
  }

  /**
   * Adds a boardListener to this board. Does nothing if <code>null</code> is passed as argument.
   * @param boardListener the boardListener to add
   */
  public void addBoardListener(final SetGameBoardListener boardListener)
  {
    _boardListeners.add(boardListener);
  }

  /**
   * Removes a listener from this board. Does nothing if <code>null</code> is passed as argument.
   * @param boardListener the listener to remove
   */
  public void removeBoardListener(SetGameBoardListener boardListener)
  {
    _boardListeners.remove(boardListener);
  }

  /**
   * Add the specified cards onto the board.
   * @param cards a non-<code>null</code> array, containing non-<code>null</code> elements.
   */
  public synchronized void addCards(final CardProperties[] cards)
  {
    for (int i = 0; i < cards.length; i++)
    {
      _board.add(cards[i]);
    }
    fireCardsAddedEvent(cards);
  }

  /**
   * @return the number of cards on the board.
   */
  public int getNbCards()
  {
    return _board.size();
  }

  /**
   * Tries to match the specified set on this board.
   * @param set the set to remove
   * @throws MatchingException if the specified set is not matching
   * @throws BoardException if one of the set card is not on the board.
   */
  public synchronized void matchSet(final CardSet set) throws MatchingException, BoardException
  {
    _logger.debug(".removeSet(" + set + "):" + CardProperties.toString(set.getCards()));
    if (!set.isMatching())
    {
      throw new MatchingException("The given set is not a matching set.");
    }

    final CardProperties[] cards = set.getCards();
    final byte notFound = -1;
    final byte[] indexes = new byte[cards.length];
    for (byte i = 0; i < cards.length; i++)
    {
      indexes[i] = notFound;
      for (byte j = 0; j < _board.size(); j++)
      {
        _logger.debug(" comparing cards[" + i + "] (" + cards[i] + ") with _board element j=" + j + "(" + (CardProperties) _board.get(j) + ")");
        if ((_board.get(j)).equals(cards[i]))
        {
          indexes[i] = j;
          _logger.debug(" found indexes[" + i + "]=" + j);
        }
      }
      if (indexes[i] == notFound)
      {
        throw new BoardException("CardProperties " + cards[i] + " not found.");
      }
    }
    for (int i = 0; i < indexes.length; i++)
    {
      _board.remove(CardProperties.findCard(cards[i]));
    }
    fireSetRemovedEvent(set);
  }

  private void fireCardsAddedEvent(final CardProperties[] cards)
  {
    if (_boardListeners != null)
    {
      for (int i = 0; i < _boardListeners.size(); i++)
      {
        final org.cb.jset.SetGameBoardListener boardListenerIF = (org.cb.jset.SetGameBoardListener) _boardListeners.get(i);
        boardListenerIF.cardsAdded(cards);
      }
    }
  }

  private void fireSetRemovedEvent(final CardSet set)
  {
    if (_boardListeners != null)
    {
      for (int i = 0; i < _boardListeners.size(); i++)
      {
        final org.cb.jset.SetGameBoardListener boardListenerIF = (org.cb.jset.SetGameBoardListener) _boardListeners.get(i);
        boardListenerIF.setRemoved(set);
      }
    }
  }
}
