// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset;

// import java.util.HashSet;

import java.io.Serializable;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Represents a card used in the game.
 * <p>
 * A card has 4 properties: a type, a color, a number of shapes and a number of fills.
 * Each property has 3 different values, making for 81 different cards in the game.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: CardProperties.java 129 2004-04-15 05:00:43Z jerome $
 */
public class CardProperties implements Serializable
{
  /**
   * @todo it is not such a good idea to hardcode the type with the name.
   * What if we change the implementation?
   */

  public static final int COLOR_RED        = 1;
  public static final int COLOR_GREEN      = 2;
  public static final int COLOR_BLUE       = 3;

  public static final int SHAPE_RECTANGLE  = 1;
  public static final int SHAPE_OVALE      = 2;
  public static final int SHAPE_ODD        = 3;

  public static final int FILL_EMPTY       = 1;
  public static final int FILL_DOTTED      = 2;
  public static final int FILL_FULL        = 3;

  private final byte _shape;

  private final byte _color;

  private final byte _number;

  private final byte _fill;

  // for test purposes
  /**
   * Creates a new instance of CardProperties, given the specified shape, color number and fill.
   * Each property has a value between <code>1</code> and <code>3</code>.
   * @param shape the card shape type
   * @param color the card color type
   * @param number the card shape number type
   * @param fill the card shape fill type
   */
  CardProperties(final int shape, final int color, final int number, final int fill)
  {
    this((byte) shape, (byte) color, (byte) number, (byte) fill);
  }

  /**
   * Creates a new instance of CardProperties, given the specified shape, color number and fill.
   * Each property has a value between <code>1</code> and <code>3</code>.
   * @param shape the card shape type
   * @param color the card color type
   * @param number the card shape number type
   * @param fill the card shape fill type
   */
  public CardProperties(final byte shape, final byte color, final byte number, final byte fill)
  {
    _shape = shape;
    _color = color;
    _number = number;
    _fill = fill;
  }

  /**
   * @return the shape property identifier
   */
  public byte getShape()
  {
    return _shape;
  }

  /**
   * @return the color property identifier
   */
  public byte getColor()
  {
    return _color;
  }

  /**
   * @return the shape number property identifier
   */
  public byte getNumber()
  {
    return _number;
  }

  /**
   * @return the shape fill property identifier
   */
  public byte getFill()
  {
    return _fill;
  }

  private static final List CARDS = createCardSet();
  /** Number of cards. **/
  public static final int CARDSNB = 81;

  private static List createCardSet()
  {
    // HashSet cards = new HashSet(CARDSNB);
    final List cards = new ArrayList(CARDSNB);
    for (byte t = 1; t < 4; t++)
    {  // type
      for (byte c = 1; c < 4; c++)
      {  // color
        for (byte n = 1; n < 4; n++)
        {  // number
          for (byte f = 1; f < 4; f++)
          {  // fill
            // cards.add(new CardProperties(t, c, n, f));
            cards.add(new CardProperties(t, c, n, f));
          }
        }
      }
    }
    return cards;
  }

  /**
   * Get the card given the specified index.
   * @param idx the index of the card.
   * @return the card matching the index.
   * @throws IllegalStateException if index is invalid.
   */
  public static CardProperties getCard(final int idx)
  {
    if (idx < 0 || idx > CARDSNB)
    {
      throw new IllegalStateException("CardProperties not found: invalid index");
    }
    return (CardProperties) CARDS.get(idx);
  }

  /**
   * @todo this should not be needed if equals is implemented correctly.
   * @param card
   * @return the CardProperties
   */
  public static CardProperties findCard(final CardProperties card)
  {
    /*
    Iterator i = CARDS.iterator();
    while (i.hasNext()) {
        CardProperties prop = (CardProperties) i.next();
        if (prop.equals(card)) {
            return prop;
        }
    }*/
    for (int i = 0; i < CARDSNB; i++)
    {
      final CardProperties prop = (CardProperties) CARDS.get(i);
      if (prop.equals(card))
      {
        return prop;
      }
    }
    throw new IllegalStateException(card + " not found in CARDS");
  }

  /*
  static CardProperties findCard(int idx) {
      Iterator i = CARDS.iterator();
      while (i.hasNext()) {
          CardProperties prop = (CardProperties) i.next();
          if (prop.equals(card)) {
              return prop;
          }
      }
      throw new IllegalStateException (card + " not found in CARDS");
  }
  */

  /**
   * Indicates whether some other object is "equal to" this one.
   * @param o the other object
   * @return <code<>true</code> if both instances are equal.
   */
  public boolean equals(final Object o)
  {
    if (this == o) return true;
    if (!(o instanceof CardProperties)) return false;

    final CardProperties cardProperties = (CardProperties) o;

    if (_color != cardProperties._color) return false;
    if (_fill != cardProperties._fill) return false;
    if (_number != cardProperties._number) return false;
    if (_shape != cardProperties._shape) return false;

    return true;
  }

  /**
   * @return a hash code value for the object.
   */
  public int hashCode()
  {
    int result;
    result = (int) _shape;
    result = 29 * result + (int) _color;
    result = 29 * result + (int) _number;
    result = 29 * result + (int) _fill;
    return result;
  }

  /**
   * An id for that card.
   * @return
   */
  private int getID()
  {
    return 27 * (getShape() - 1)
          + 9 * (getColor() - 1)
          + 3 * (getNumber() - 1)
          + 1 * (getFill() - 1);
  }

  /**
   * @return a <code>String</code> representation of that object.
   */
  public String toString()
  {
    return "" + getID();
  }

  /**
   * A helper function that generates a <code>String</code> representation for the
   * specified array of <code>CardProperties</code>.
   * @param cards the cards for which we want a <code>String</code> representation.
   * @return the <code>String</code> representation
   */
  public static String toString(final CardProperties[] cards)
  {
    return toString(Arrays.asList(cards));
  }

  /**
   * A helper function that generates a <code>String</code> representation for the
   * specified list of <code>CardProperties</code>.
   * @param cards the cards for which we want a <code>String</code> representation.
   * @return the <code>String</code> representation
   */
  public static String toString(final List cards)
  {
    final StringBuffer buf = new StringBuffer("[");
    if (cards != null)
    {
      for (int i = 0; i < cards.size(); i++)
      {
        final org.cb.jset.CardProperties cardProperties = (org.cb.jset.CardProperties) CARDS.get(i);
        if (i != 0)
        {
          buf.append(",");
        }
        if (cardProperties != null)
        {
          buf.append(cardProperties.getID());
        }
        else
        {
          buf.append("null");
        }
      }
    }
    buf.append("]");
    return buf.toString();
  }
}
