// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset.client;

import org.cb.jset.server.RemoteSetGame;
import org.cb.jset.server.RemoteSetGameServer;
import org.cb.jset.server.RemoteSetGamePlayer;
import org.cb.jset.server.RemoteSetGameConnection;
import org.cb.jset.SetGamePlayer;
import org.cb.jset.CardProperties;
import org.cb.jset.CardSet;
import org.cb.jset.SetGameConnection;
import org.cb.jset.MatchingException;
import org.cb.jset.SetGame;

import java.rmi.Naming;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.NotBoundException;
import java.rmi.server.UnicastRemoteObject;
import java.io.Serializable;
import java.net.MalformedURLException;

/**
 * FIXME rename. More like a RemoteGameHelper class
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: Client.java 129 2004-04-15 05:00:43Z jerome $
 */
public class Client
{
  private static final org.apache.log4j.Logger LOGGER = org.apache.log4j.Logger.getLogger("Client");

  static class LocalPlayerWrapper extends UnicastRemoteObject implements Serializable, RemoteSetGamePlayer
  {
    private final transient SetGamePlayer _localPlayer;

    public LocalPlayerWrapper(final SetGamePlayer localPlayer)
        throws RemoteException
    {
      _localPlayer = localPlayer;
    }

    public String getName() throws RemoteException
    {
      return _localPlayer.getName();
    }

    public void block() throws RemoteException
    {
      _localPlayer.block();
    }

    public void unblock() throws RemoteException
    {
      _localPlayer.unblock();
    }

    public boolean isBlocked() throws RemoteException
    {
      return _localPlayer.isBlocked();
    }

    public void disconnect() throws RemoteException
    {
      _localPlayer.disconnect();
    }

    public void cardsAdded(final CardProperties[] cards) throws RemoteException
    {
      _localPlayer.cardsAdded(cards);
    }

    public void setRemoved(final CardSet set) throws RemoteException
    {
      _localPlayer.setRemoved(set);
    }
  }

  static class RemoteConnectionWrapper implements SetGameConnection
  {
    private final RemoteSetGameConnection _remoteConnection;
    private boolean _wasDisconnected;
    private RemoteException _remoteException;

    void treatConnectionProblem(final RemoteException e)
    {
      _wasDisconnected = true;
      _remoteException = e;
      throw new IllegalStateException("Failure to make a remote call... " + e.getMessage());
    }

    public RemoteConnectionWrapper(final RemoteSetGameConnection localConnection)
    {
      _remoteConnection = localConnection;
    }

    public void matchSet(final CardSet cards) throws MatchingException
    {
      try
      {
        _remoteConnection.matchSet(cards);
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
    }

    public void close()
    {
      try
      {
        _remoteConnection.close();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
    }

    public boolean isClosed()
    {
      try
      {
        return _remoteConnection.isClosed();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return false; // please compiler
    }
  }


  static class RemoteGameWrapper implements SetGame
  {
    private final RemoteSetGame _remoteGame;
    private boolean _wasDisconnected;
    private RemoteException _remoteException;

    void treatConnectionProblem(final RemoteException e)
    {
      _wasDisconnected = true;
      _remoteException = e;
      throw new IllegalStateException("Failure to make a remote call... " + e.getMessage());
    }

    public RemoteGameWrapper(final RemoteSetGame remoteGame)
    {
      _remoteGame = remoteGame;
    }

    public int getID()
    {
      try
      {
        return _remoteGame.getID();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return -1; // please compiler
    }

    public String getName()
    {
      try
      {
        return _remoteGame.getName();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return ""; // please compiler
    }

    public void start()
    {
      // FIXME
    }

    public SetGameConnection connect(final SetGamePlayer player)
    {
      try
      {
        final RemoteSetGamePlayer remotePlayer = new LocalPlayerWrapper(player);
        return new RemoteConnectionWrapper(_remoteGame.connect(remotePlayer));
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return null; // please compiler
    }

    public int getMaxNbPlayers()
    {
      try
      {
        _remoteGame.getMaxNbPlayers();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return -1; // please compiler
    }

    public int getCurrentNbPlayers()
    {
      try
      {
        _remoteGame.getCurrentNbPlayers();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return -1; // please compiler
    }
  }


  public static void main(final String[] args)
  {
    org.apache.log4j.PropertyConfigurator.configure("log4j.properties");
    String host = "localhost";
    String gameName = "TestGame";
    String playerName = "TestGame";
    final int nbParsedArgs = Math.min(3, args.length);
    switch (nbParsedArgs)
    {
      case 3:
        playerName = args[2];
      case 2:
        gameName = args[1];
      case 1:
        host = args[0];
    }
    if (System.getSecurityManager() == null)
    {
      System.setSecurityManager(new RMISecurityManager());
    }
    try
    {
      final RemoteSetGame game = getGame(host, gameName);

      if (game != null)
      {
        LOGGER.info("Game getName " + game.getName());

//        SimplePlayer player = new SimplePlayer(playerName);
        final SmartAutoPlayer player = new SmartAutoPlayer(playerName);
        player.startPlayingGame(game);
      }
    }
    catch (Exception e)
    {
      LOGGER.error("Client Error: " + e.getMessage(), e);
    }
  }

  public static RemoteSetGame getGame(final String host, final String gameName)
      throws NotBoundException, MalformedURLException, RemoteException
  {
    final String name = "//" + host + "/JSetServer";
//            Object obj = Naming.lookup(getName);
//            logger.info("Server of type:" + obj.getClass());
    final RemoteSetGameServer jSetServer = (RemoteSetGameServer) Naming.lookup(name);
    final RemoteSetGame[] games = jSetServer.getGames();
    LOGGER.info(games.length + " games found on the server " + name);
    for (int i = 0; i < games.length; i++)
    {
      LOGGER.info(games[i] + " <" + games[i].getName() + "> " + games[i].getMaxNbPlayers());
    }

    RemoteSetGame game = null;
    final int nbPlayers = 2;
    // find a game with 2 players
    for (int i = 0; i < games.length; i++)
    {
      if (games[i].getName().equals(gameName))
      {
//                    == nbPlayers &&
//                games[i].getMaxNbPlayers() == nbPlayers &&
//                    games[i].getCurrentNbPlayers() < nbPlayers) {
        LOGGER.info("Found a game:" + games[i] + " <" + games[i].getName() + "> " + games[i].getMaxNbPlayers());
        game = games[i];
        break;
      }
    }
    if (game == null)
    {
      LOGGER.info("Creating a game: getName:" + gameName + " nbplayers " + nbPlayers);
      final int gameID = jSetServer.createGame(gameName, nbPlayers);
      LOGGER.info("Game created with ID:" + gameID);
      game = jSetServer.getGame(gameID);
    }
    return game;
  }
}
