// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset.client;

import org.apache.log4j.Logger;
import org.cb.jset.*;
import org.cb.jset.server.RemoteSetGame;

import java.util.List;
import java.util.ArrayList;

/**
 * A player that finds sets automatically.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: SmartAutoPlayer.java 123 2004-04-14 23:45:53Z jerome $
 */
public class SmartAutoPlayer implements SetGameListener
{
  private Logger _logger;
  private SetFinderThread _thread;
  private MatchingSetFinder _finder;
  private String _name;
  private NetworkGame _game;
  private BoardImpl _board = new BoardImpl();

  private final GameController _gameController = new GameController();

  /**
   * Creates a new instance with the specified player playerName.
   */
  public SmartAutoPlayer(final String playerName)
  {
    _logger = Logger.getLogger("SmartAutoPlayer-" + playerName);
    _name = playerName;
    _finder = new MatchingSetFinder();
    _thread = new SetFinderThread();
  }

  public String name()
  {
    return _name;
  }

  private void printCards(final String msg, final CardProperties[] cards)
  {
    if (_logger.isInfoEnabled())
    {
      final StringBuffer buf = new StringBuffer(msg);
      buf.append(CardProperties.toString(cards));
      _logger.info(buf.toString());
    }
  }

  void startPlayingGame(final RemoteSetGame game)
  {
    final NetworkGame networkGame = new NetworkGame("SmartAutoPlayerUser");
    networkGame.setRemoteGame(game);
    _game = networkGame;
    _game.addBoardListener(_gameController);
    _game.addGameListener(_gameController);
    _game.start();
    _thread.start();
  }


  /**
   * Control events coming from the game instance.
   * 
   * FIXME we probably need to add multithreading handling on the model.
   */
  class GameController implements SetGameBoardListener, SetGameListener
  {
    /**
     * @inheritDoc
     */
    public void cardsAdded(final CardProperties[] cards)
    {
      _board.addCards(cards);
    }

    /**
     * @inheritDoc
     */
    public void setRemoved(final CardSet set)
    {
      try
      {
        _board.matchSet(set);
      }
      catch (Exception e)
      {
        throw new IllegalStateException("BoardException should not happen here: " + e.getMessage());
      }
    }

    /**
     * @inheritDoc
     */
    public void gameChange(final GameEvent event)
    {
      // FIXME
    }
  }

  // FIXME: not sure if we really need that.
  public synchronized void gameChange(final GameEvent event)
  {
    if (event.getType() == GameEvent.END)
    {
      notifyAll();
    }
  }

  class SetFinderThread extends Thread
  {
    public void run()
    {
      play();
    }
  }

  private void play()
  {
    while (! _game.wasDisconnected())
    {
      tryToFindSet();
    }
  }

  void tryToFindSet()
  {
    if (_logger.isInfoEnabled())
    {
      _logger.info("-- cards added - starting to find --");
    }
    int sleepTime = (int) (Math.random() * 6000);
    if (_logger.isInfoEnabled())
    {
      _logger.info("Sleep for " + sleepTime + "ms.");
    }
    try
    {
      synchronized (SmartAutoPlayer.this)
      {
        wait(sleepTime);
      }
    }
    catch(InterruptedException e)
    {
    }

    if (true) // only for human opponent.
    {
      sleepTime = 60000;
      try
      {
        Thread. sleep(sleepTime);
      }
      catch (InterruptedException e)
      {
      }
    }

    if (_logger.isInfoEnabled())
    {
      _logger.info("-- Awaken! -- ");
    }
    if (_game.wasDisconnected())
    {
      _logger.info("-- Connection was closed in the meantime -- ");
      return;
    }
    if (_logger.isInfoEnabled())
    {
      _logger.info("-- Searching for sets -- ");
    }
    printCards("-- Cards on BoardImpl:", _board.getCards());

    final List elements = new ArrayList();
    for (int i = 0; i < _board.getCards().length; i++)
    {
      final CardProperties cardProperties = _board.getCards()[i];
      elements.add(new JSetCard(cardProperties));
    }
    final CardSet[] sets = _finder.findSets(elements);
    printMatchingSets(sets);
    if (sets.length > 0)
    {
      final int randIdx = (int) (sets.length * Math.random());
      if (_logger.isInfoEnabled())
      {
        _logger.info("Trying to remove set Idx" + randIdx + ": " + sets[randIdx]);
      }
      try
      {
        _game.removeSet(sets[randIdx]);
      }
      catch (MatchingException e)
      {
        if (_logger.isInfoEnabled())
        {
          _logger.info("Matching not accepted!", e);
        }
      }
    }
        // we should be smarter here as some cards may be shared within several sets
        // we risk to be blocked!
//                for (int i = 0; i < sets.length; i++) {
//                    _connection.removeSet(sets[i]);
//                }
//            }

  }

  void printMatchingSets(final CardSet[] foundSets)
  {
    if (_logger.isInfoEnabled())
    {
      final int maxPrinted = 5;
      _logger.info("Found " + foundSets.length + " sets.");
      int i;
      for (i = 0; i < foundSets.length && i < maxPrinted; i++)
      {
        _logger.info("CardSet[" + i + "]:" + foundSets[i]);
      }
      if (i < foundSets.length)
      {
        _logger.info("Skipping " + (foundSets.length - i) + " set(s)...");
      }
    }
  }
}
