// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset.client.model;

import org.apache.log4j.Logger;
import org.cb.jset.JSetCard;
import org.cb.jset.CardProperties;
import org.cb.jset.CardSet;
import org.cb.jset.BoardException;
import org.cb.cardboard.AbstractCardBoardModel;
import org.cb.cardboard.Card;
import org.cb.cardboard.CardBoardEvent;

import java.util.Collections;
import java.util.List;
import java.util.ArrayList;
import java.io.Serializable;


/**
 * A client BoardImpl is a special board that also keeps track of card layout, especially the empty spaces.
 *
 *
 * Note: implementation not yet multithread safe (FIXME)
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: JSetClientBoardModel.java 129 2004-04-15 05:00:43Z jerome $
 */
public class JSetClientBoardModel extends AbstractCardBoardModel implements SetGameClientBoard, Serializable
{
  private transient final Logger _logger = Logger.getLogger(JSetClientBoardModel.class);

  /** A List of JSetCard elements representing the cards on the board. */
  private final List _cards = new ArrayList();

  /**
   * Clear the board.
   */
  public void clear()
  {
    int size = _cards.size();
    _cards.clear();
//    _selectedCards.clear();
    fireIntervalRemoved(0, size - 1);
  }

  /**
   * @return the number of spaces used by cards on the board. Unused spaces are taken into account.
   * @see #getNbCards() for the exact number of cards on the board.
   */
  public int getSize()
  {
    return _cards.size();
  }

  /**
   * Return the Card element specified by the index on the board.
   * @param index
   * @return the Card element or <code>null</code> if the space on the board is empty at that position.
   */
  public Card getElementAt(final int index)
  {
    return (Card) _cards.get(index);
  }

  /**
   * @return The cards in the board. May contain some <code>null</code> to represent the empty spaces.
   */
  public List getElements()
  {
    return Collections.unmodifiableList(_cards);
  }

  /**
   * @return the number of cards on the board.
   * @see #getSize() which takes into account the number of cards plus the empty spaces.
   */
  public int getNbCards()
  {
    int lNbCards = 0;
    for (int i = 0; i < _cards.size(); i++)
    {
      final Card card = (Card) _cards.get(i);
      if (card != null)
      {
        lNbCards++;
      }
    }
    return lNbCards;
  }

  /**
   * Add the specified cards onto the board.
   * @param cards the cards to add.
   * @throws NullPointerException if <code>cards</code> is <code>null</code> or if an element in the array is <code>null</code>.
   */
  public void addCards(final CardProperties[] cards)
  {
    logCards("before adding: " + CardProperties.toString(cards));

    int indexOfCardsToAdd = 0;
    int indexInCardsList = 0; // FIXME we could cache this one.


    final List indexesOfCardsChanged = new ArrayList();
    cardsToAdd:
    // card.addCardListener(_eventHandler);
    for (; indexInCardsList < _cards.size() && indexOfCardsToAdd < cards.length; indexInCardsList++)
    {
      final JSetCard oldCard = (JSetCard) _cards.get(indexInCardsList);
      if (oldCard == null)
      {
        final CardProperties cardProperties = cards[indexOfCardsToAdd++];
        final Card newCard = new org.cb.jset.JSetCard(cardProperties);
        _cards.set(indexInCardsList, newCard);
        indexesOfCardsChanged.add(new Integer(indexInCardsList));
        continue cardsToAdd;
      }
    }
    if (indexesOfCardsChanged.size() > 0)
    {
      fireContentsChanged(indexesOfCardsChanged);
    }
    if (indexOfCardsToAdd < cards.length)
    {
      final int firstListIndex = indexInCardsList;
      int lastListIndex = 0;
      for (; indexOfCardsToAdd < cards.length; indexOfCardsToAdd++)
      {
        final CardProperties cardProperties = cards[indexOfCardsToAdd];
        final Card card = new org.cb.jset.JSetCard(cardProperties);
        _cards.add(indexInCardsList++, card);
        lastListIndex = indexInCardsList;
      }
      fireIntervalAdded(firstListIndex, (lastListIndex - 1));
    }
    logCards("after adding: " + CardProperties.toString(cards));
  }

  /**
   * Remove the specified matching set from the board.
   * @param set the set to remove
   * @throws BoardException when something fails
   */
  public void removeSet(final CardSet set) throws BoardException
  {
    logCards("before matching: "  + CardProperties.toString(set.getCards()));
    final CardProperties[] cards = set.getCards();

    final int[] indexesOnBoard = new int[cards.length];
    for (int i = 0; i < indexesOnBoard.length; i++)
    {
      final int index = _cards.indexOf(new org.cb.jset.JSetCard(CardProperties.findCard(cards[i])));
      if (index == -1)
      {
        _logger.debug("JSetCard " + cards[i] + " not found. Listing contents of board...");
        logCards("");
        throw new BoardException("JSetCard " + cards[i] + " not found.");
      }
      indexesOnBoard[i] = index;
    }

    for (int i = 0; i < indexesOnBoard.length; i++)
    {
      final int lIndex = indexesOnBoard[i];
      _cards.set(lIndex, null);
      fireIntervalRemoved(lIndex, lIndex);
    }
    logCards("after matching: "  + CardProperties.toString(set.getCards()));
  }

  private void logCards(final String message)
  {
    _logger.debug("Logging cards: " + message);
    for (int j = 0; j < _cards.size(); j++)
    {
      final org.cb.jset.JSetCard card = (org.cb.jset.JSetCard) _cards.get(j);
      _logger.debug("[" + j + "]:" + card);
    }
  }

  private void fireContentsChanged(final List indexes)
  {
    final CardBoardEvent event = new CardBoardEvent(this, indexes);
    fireContentsChanged(event);
  }

  private void fireIntervalAdded(final int firstIndex, final int lastIndex)
  {
    final CardBoardEvent event = new CardBoardEvent(this, CardBoardEvent.INTERVAL_ADDED, firstIndex, lastIndex);
    fireIntervalAdded(event);
  }

  private void fireIntervalRemoved(final int firstIndex, final int lastIndex)
  {
    final CardBoardEvent event = new CardBoardEvent(this, CardBoardEvent.INTERVAL_REMOVED, firstIndex, lastIndex);
    fireIntervalRemoved(event);
  }

//  class EventHandler implements SetGameClientCardListener
//  {
//    public void notifyCardEvent(CardEvent card)
//    {
//      if (card.getType() == CardEvent.SELECTED)
//      {
//        _selectedCards.add(card.getSource());
//      }
//      else
//      {
//        _selectedCards.remove(card.getSource());
//      }
//    }
//  }
}      
