// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset.client.ui;

import org.cb.jset.CardProperties;
import org.cb.jset.JSetCard;
import org.cb.jset.client.model.SetGameClientCardListener;
import org.cb.jset.client.model.CardEvent;
import org.apache.log4j.Logger;

import javax.swing.JPanel;
import javax.swing.JComponent;
import javax.swing.BorderFactory;
import javax.swing.border.Border;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.Paint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.TexturePaint;
import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.geom.RectangularShape;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.Map;
import java.io.Serializable;

/**
 * The visual component representing a card on the board.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: JSetCardComponent.java 130 2004-04-15 05:18:07Z jerome $
 */
public class JSetCardComponent extends JPanel
{
  private final transient Logger _logger;
  private ShapeComponent[] _shapes;

  private boolean _isSelected = false;
  // private Border _selectedBorder = BorderFactory.createLoweredBevelBorder();
  private Border _selectedBorder = BorderFactory.createLineBorder(Color.black, 2);
  private Border _unselectedBorder = BorderFactory.createRaisedBevelBorder();
  private EventHandler _eventHandler = new EventHandler();
  private JSetCard _card;

  /**
   * A class representing the third type of shape.
   * @todo make it a subclass of RectangularShape ?
   */
  class OddShape
  {
    // int xa, ya, x1, y1, x2, y2, x3, y3;
    private final GeneralPath _oddShape;
//    private boolean _firstTime;
//    private final Rectangle _area;
    /**
     * Constructs an instance of <code>OddShape</code>.
     */
    public OddShape()
    {
      _oddShape = new GeneralPath();
//      _firstTime = true;
//      _area = new Rectangle();
    }

    /**
     * Return a general path that walks throught the contour of the odd shape, given the specified bounds.
     * @param bounds
     * @return the general path of the odd shape
     */
    GeneralPath constructOddShape(final Rectangle bounds)
    {
      _logger.debug("building odd shape " + bounds);
      //
      //   x0,y0 --- x1,y1
      //     \         \
      //      \         \
      //      |         |
      //     /         /
      //    /         /
      //   |         |
      //    \         \
      //   x2,y2 ---- x3,y3

      final int margin = 8;

      final int x0 = (int) bounds.getX() + margin;
      final int y0 = (int) bounds.getY();
      _oddShape.moveTo(x0, y0);
      final int x1 = x0 + (int) bounds.getWidth() - 2 * margin;
      final int y1 = y0;
      _oddShape.lineTo(x1, y1);

      final int x3 = x1;
      final int y3 = y1 + (int) bounds.getHeight();
      final int bez1x1 = x1 + 15;
      final int bez1y1 = y1 + 10;
      final int bez1x2 = x1 - 15;
      final int bez1y2 = y3 - 10;
      _oddShape.curveTo(bez1x1, bez1y1, bez1x2, bez1y2, x3, y3);

      final int x2 = x0;
      final int y2 = y3;
      _oddShape.lineTo(x2, y2);

      // temp
      final int bez2x1 = x2 - 15;
      final int bez2y1 = y2 - 10;
      final int bez2x2 = x2 + 15;
      final int bez2y2 = y0 + 10;
      _oddShape.curveTo(bez2x1, bez2y1, bez2x2, bez2y2, x0, y0);
//      _oddShape.lineTo(x0, y0);

      return _oddShape;
    }
  }

  /**
   * A graphical component representing the shapes on the card.
   */
  final class ShapeComponent extends JComponent
  {
    private final Color _color;
    private final CardProperties _cardProperties;
    private final transient Logger _logger;
    private Map _shapes;

    /**
     * Constructs an instance of <code>ShapeComponent</code> given the specified card properties, and index
     * on the card.
     * @param cardProperties the properties used to determine the shape
     * @param index the index used to determine the position.
     */
    private ShapeComponent(final CardProperties cardProperties, final int index)
    {
      _cardProperties = cardProperties;
      _color = getColor(_cardProperties.getColor());
      _logger = Logger.getLogger("Shape " + index + " " + cardProperties);
      setBackground(Color.white);
    }

    private Color getColor(final byte pColor)
    {
      switch(pColor)
      {
        case CardProperties.COLOR_RED:
          return Color.RED;
        case CardProperties.COLOR_GREEN:
          return new Color(0, 102, 0);
        case CardProperties.COLOR_BLUE:
          return Color.BLUE;
        default:
          throw new IllegalStateException("Invalid color " + pColor);
      }
    }

    /**
     * @inheritDoc
     */
    protected void paintComponent(final Graphics g)
    {
      _logger.debug("painting shape component: " +  getBounds());
      super.paintComponent(g);    //To change body of overridden methods use File | Settings | File Templates.

      final Graphics2D g2 = (Graphics2D) g;

      _logger.debug("this bounds: " +  getBounds());
      _logger.debug("container bounds: " +  getParent().getBounds());

      final Paint oldPaint = g2.getPaint();

      switch (_cardProperties.getFill())
      {
        case CardProperties.FILL_EMPTY :
        case CardProperties.FILL_FULL :
          g2.setPaint(_color);
          break;
        case CardProperties.FILL_DOTTED:
          final BufferedImage bi = new BufferedImage(5, 5, BufferedImage.TYPE_INT_RGB);
          final Graphics2D big = bi.createGraphics();
          big.setColor(_color);
          big.fillRect(0, 0, 5, 5);
          big.setColor(getParent().getBackground());
          big.fillOval(0, 0, 5, 5);
          final Rectangle r = new Rectangle(0,0,5,5);
          g2.setPaint(new TexturePaint(bi, r));
          break;
      }

      // set the stroke
      switch (_cardProperties.getFill())
      {
        case CardProperties.FILL_EMPTY:
          final Stroke biggerStroke = new BasicStroke(4.0f);
          g2.setStroke(biggerStroke);
          break;
        case CardProperties.FILL_FULL :
          break;
        case CardProperties.FILL_DOTTED:
          break;
      }

      final Shape shape = getShape(_cardProperties.getShape());
      _logger.debug("%%% shape: " + shape + " bounds " + shape.getBounds());

      // Determines whether to fill, stroke, or fill and stroke.
      switch (_cardProperties.getFill())
      {
        case CardProperties.FILL_FULL :
          g2.fill(shape);
          break;
        case CardProperties.FILL_DOTTED:
          final Graphics2D tempg2 = g2;
          g2.fill(shape);
          g2.setColor(Color.darkGray);
          g2.draw(shape);
          g2.setPaint(tempg2.getPaint());
          break;
        case CardProperties.FILL_EMPTY:
          g2.draw(shape);
          break;
      }

      _logger.debug("painting circle: " +  getSize());

      g2.setPaint(oldPaint);
    }

    private Map initShapes()
    {
      final Map shapes = new HashMap(3);

      final Rectangle rectangle = new Rectangle();
      final Rectangle internRectangle = new Rectangle();
      shapes.put(new Integer(CardProperties.SHAPE_RECTANGLE), createShape(rectangle, internRectangle));

      final Ellipse2D.Double oval = new Ellipse2D.Double();
      final Ellipse2D.Double internOval = new Ellipse2D.Double();
      shapes.put(new Integer(CardProperties.SHAPE_OVALE), createShape(oval, internOval));

      final Rectangle lShapeBounds =
          new Rectangle(0, 0, (int) getBounds().getWidth(), (int) getBounds().getHeight());
      final Shape odd = new OddShape().constructOddShape(lShapeBounds);
      shapes.put(new Integer(CardProperties.SHAPE_ODD), odd);

      return shapes;
    }

    /**
     * Create a shape by {@link Area#subtract(java.awt.geom.Area) substracting} the specified
     * internal shape to the specified external one.
     * @param externalShape
     * @param internShape
     * @return
     */
    private Area createShape(final RectangularShape externalShape, final RectangularShape internShape)
    {
      externalShape.setFrame(0, 0, getBounds().getWidth(), getBounds().getHeight());
      // FIXME rename
      final int xxxX = (int) (getBounds().getWidth() / 4);
      final int xxxY = (int) (getBounds().getHeight() / 4);
      internShape.setFrame(xxxX, xxxY, getBounds().getWidth() - 2 * xxxX, getBounds().getHeight() - 2 * xxxY);
      final Area area2 = new Area(externalShape);
      final Area internOv = new Area(internShape);
      area2.subtract(internOv);
      return area2;
    }

    private Shape getShape(final byte shapeID)
    {
      // lazily construct the shapes so that they are built at a time when surrouding size information is known.
      if (_shapes == null)
      {
        _shapes = this.initShapes();
      }
      return (Shape) _shapes.get(new Integer(shapeID));
    }
  }

  /**
   * Constructs an instance of <code>JSetCardComponent</code> given the specified card.
   * @param card
   */
  public JSetCardComponent(final JSetCard card)
  {
    _logger = Logger.getLogger("Shape " + card);
    _card = card;
    _shapes = initShapes(card.getProperties());
    for (int i = 0; i < _shapes.length; i++)
    {
      final ShapeComponent mShape = _shapes[i];
      this.add(mShape);
    }
    init();
  }

  private void init()
  {
    this.setBackground(Color.white);
    this.addMouseListener(_eventHandler);
    this.setBorder(_unselectedBorder);
    final int width = 80;
    final int height = (int) (1.5 * width);
    setSize(width, height);
    setPreferredSize(getSize());
    _logger.debug("initialized: " +  getSize());
  }

  private ShapeComponent[] initShapes(final CardProperties pProperties)
  {
    final ShapeComponent[] lResult = new ShapeComponent[pProperties.getNumber()];
    for (int i = 0; i < lResult.length; i++)
    {
      lResult[i] = new ShapeComponent(pProperties, i);
    }
    return lResult;
  }


  private void resetShapeBounds()
  {
    final int width = (int) this.getBounds().getWidth();
    final int height = (int) this.getBounds().getHeight();
    // margins when displaying 3 shapes
    final int sideMargin = 5;
    final int middleMargin = 5;
    final int shapeWidth = (width - 2 * sideMargin - middleMargin) / 2;
    final int shapeHeight = (height - 2 * sideMargin - middleMargin) / 2;
    _logger.debug("** setting new shapes bounds: " + this.getBounds());
    _logger.debug("** shapeWidth: " + shapeWidth + " shapeHeight: " + shapeHeight);

    switch (_shapes.length)
    {
      case 1:
        _shapes[0].setBounds((width - shapeWidth) / 2, (height - shapeHeight) / 2, shapeWidth, shapeHeight);
        break;
      case 2:
        // upper _area
        _shapes[0].setBounds((width - shapeWidth) / 2, sideMargin, shapeWidth, shapeHeight);
        // lower _area
        _shapes[1].setBounds((width - shapeWidth) / 2, sideMargin + shapeHeight + middleMargin, shapeWidth, shapeHeight);
        break;
      case 3:
        // upper _area
        _shapes[0].setBounds(sideMargin, sideMargin, shapeWidth, shapeHeight);
        _shapes[1].setBounds(sideMargin + shapeWidth + middleMargin, sideMargin, shapeWidth, shapeHeight);
        // lower _area
        _shapes[2].setBounds((width - shapeWidth) / 2, sideMargin + shapeHeight + middleMargin, shapeWidth, shapeHeight);
        break;
      default:
        throw new IllegalStateException("Invalid number of shapes " + _shapes.length);
    }
    for (int i = 0; i < _shapes.length; i++)
    {
      final ShapeComponent mShape = _shapes[i];
      _logger.debug("shape[" + i + "]s bounds: " + mShape.getBounds());
    }
  }

  // debugging code
  /*
  public void paint(final Graphics g)
  {
    super.paint(g);    //To change body of overridden methods use File | Settings | File Templates.
    _logger.debug("painting card: " +  getBounds());
  }

  // debugging code
  public void setSize(int width, int height)
  {
    _logger.debug("setting size, " + width + " " + height);
    super.setSize(width, height);    //To change body of overridden methods use File | Settings | File Templates.
  }

  public void setBounds(final int x, final int y, final int width, final int height)
  {
    _logger.debug("setting bounds, " + x + " " + y + " " + width + " " + height);
    super.setBounds(x, y, width, height);
    _logger.debug("this bounds: " +  getBounds());
    if (getParent() != null)
    {
      _logger.debug("container bounds: " +  getParent().getBounds());
    }
  }
  */

  /**
   * Update the border depending on the internal selection status.
   * @see #isSelected
   */
  public void updateBorder()
  {
    if (_isSelected)
    {
      setBorder(_selectedBorder);
    }
    else
    {
      setBorder(_unselectedBorder);
    }
    invalidate();
  }

  /**
   * Modifies the internal selection status
   */
  private void changeSelection()
  {
    setSelection(! _isSelected);
  }

  /**
   * Set the new internal selection status for the card.
   * If the selection has changed, fire an new {@link CardEvent} with the appropriate event type.
   * @param newSelection
   * @see #isSelected()
   */
  public void setSelection(final boolean newSelection)
  {
    final boolean oldSelection = _isSelected;
    if (oldSelection != newSelection)
    {
      _isSelected = newSelection;
      final int selected = _isSelected ? CardEvent.SELECTED : CardEvent.DESELECTED;
      fireCardSelectionEvent(new CardEvent(getCard(), selected));
    }
  }

  /**
   * @return <code>true</code> if the card is selected, <code>false</code> otherwise.
   */
  public boolean isSelected()
  {
    return _isSelected;
  }

  /**
   * @return the card
   */
  public JSetCard getCard()
  {
    return _card;
  }

  /**
   * Paint this card component given the specified graphics.
   * @param g
   */
  protected void paintComponent(final Graphics g)
  {
    _logger.debug("painting card component: " +  getBounds());
    super.paintComponent(g);    //To change body of overridden methods use File | Settings | File Templates.
    resetShapeBounds();
    updateBorder();
  }

  /**
   * @param i
   * @return the ith shape from that card.
   */
  ShapeComponent getShape(int i)
  {
    return _shapes[i];
  }

  /**
   * A mouse listener that {@link JSetCardComponent#changeSelection changes the card's selection} upon a mouse click.
   */
  private class EventHandler extends MouseAdapter implements Serializable
  {
    /**
     * @inheritDoc
     */
    public void mouseClicked(final MouseEvent e)
    {
      // Checks whether or not the cursor is inside of the rectangle when the
      // user releases the mouse button.
			if(JSetCardComponent.this.contains(e.getX(), e.getY()))
      {
        changeSelection();
      }
    }
  }

  /**
   * Adds the specified cardlistener to the component listener list.
   * @param cardListener
   */
  void addCardEventListener(final SetGameClientCardListener cardListener)
  {
    super.listenerList.add(SetGameClientCardListener.class, cardListener);
  }

  /**
   * Removes the specified cardlistener to the component listener list.
   * @param cardListener
   */
  void removeCardEventListener(final SetGameClientCardListener cardListener)
  {
    super.listenerList.remove(SetGameClientCardListener.class, cardListener);
  }

  /** fire the specified card selection event
   * @param e the event to fire
   * @see SetGameClientCardListener#notifyCardEvent(org.cb.jset.client.model.CardEvent)
   */
  private void fireCardSelectionEvent(final CardEvent e)
  {
    // Guaranteed to return a non-null array
    final Object[] listeners = super.listenerList.getListenerList();
    _logger.debug("" + listeners.length);
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length - 2; i >= 0; i -= 2)
    {
      if (listeners[i] == SetGameClientCardListener.class)
      {
        ((SetGameClientCardListener) listeners[i + 1]).notifyCardEvent(e);
      }
    }
  }
}
