// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset.client.ui;

import javax.swing.border.Border;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.BorderFactory;
import javax.swing.JMenuItem;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.HeadlessException;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.util.Iterator;
import java.util.Map;
import java.util.HashSet;
import java.util.HashMap;
import java.rmi.RMISecurityManager;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.net.MalformedURLException;
import java.io.Serializable;

import org.apache.log4j.Logger;
import org.cb.jset.client.model.JSetClientBoardModel;
import org.cb.jset.server.RemoteSetGame;
import org.cb.jset.client.model.SetGameClientCardListener;
import org.cb.jset.client.model.CardEvent;
import org.cb.jset.client.SetGameListener;
import org.cb.jset.client.GameEvent;
import org.cb.jset.client.LocalGame;
import org.cb.jset.client.NetworkGame;
import org.cb.jset.client.Client;
import org.cb.jset.CardProperties;
import org.cb.jset.CardSet;
import org.cb.jset.MatchingSetFinder;
import org.cb.jset.JSetCard;
import org.cb.jset.SetGameBoardListener;
import org.cb.jset.MatchingException;
import org.cb.jset.JSetBuild;
import org.cb.jset.BoardException;
import org.cb.cardboard.CardBoardEvent;
import org.cb.cardboard.CardBoardModelListener;

/**
 * The main UI for the game.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: JSetClientUI.java 128 2004-04-15 04:12:04Z jerome $
 */
public class JSetClientUI extends JFrame
{
  final transient Logger _logger = Logger.getLogger("JSetClientUI");

  private JFrame _frame = this;

  private JSetBoardComponent _boardPanel;
  private final JSetClientBoardModel _boardModel = new JSetClientBoardModel();
  private final java.util.Set _selectedCards = new HashSet();
  private final Map _mapCardsToComponents = new HashMap();

  private final BoardController _boardController = new BoardController();
  private final GameController _gameController = new GameController();
//  private CardStack _cardStack;
  private JLabel _infoLabel;

  private org.cb.jset.client.SetGame _game;

  /** Wether we ask to confirm close.
   * @todo make it configurable
   */
  private boolean _askCloseConfirm = true;

  /**
   * Constructs an instance of JSetClientUI, given the specified title.
   * @param title the title to put on the UI.
   * @throws HeadlessException if GraphicsEnvironment.isHeadless() returns true.
   */
  public JSetClientUI(final String title)
      throws HeadlessException
  {
    super(title);
    init();
  }

  // debug code.
  /**
  public void paint(final Graphics g)
  {
    super.paint(g);    //To change body of overridden methods use File | Settings | File Templates.
    _logger.debug("painting frame: " +  getSize());
  }

  public void paintComponents(final Graphics g)
  {
    super.paintComponents(g);    //To change body of overridden methods use File | Settings | File Templates
    _logger.debug("painting component: " +  getSize());
  }
  **/

  /**
   * Construct the UI.
   */
  private void init()
  {
    final JPanel panel = new JPanel(new BorderLayout());
    _boardPanel = new JSetBoardComponent(_boardModel);
    final Border etchedBorder = BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Card Board");
    _boardPanel.setSize(new Dimension(800, 600));
    _boardPanel.setBorder(etchedBorder);
    _boardPanel.setLayout(new FlowLayout(FlowLayout.LEADING, 20, 20));

    _boardModel.addBoardDataListener(_boardController);
//    _boardPanel.addBoardSelectionListener(_boardController);

    _infoLabel = new JLabel();

    final JMenuBar menuBar;
    JMenu menu;
    JMenuItem menuItem;

    menuBar = new JMenuBar();
    menu = new JMenu("Edit");
    menu.setMnemonic(KeyEvent.VK_E);
    menu.getAccessibleContext().setAccessibleDescription("Edit options");
    menuBar.add(menu);

    menu = new JMenu("View");
    menu.setMnemonic(KeyEvent.VK_V);
    menu.getAccessibleContext().setAccessibleDescription("View");
    menuBar.add(menu);

    menu = new JMenu("Game");
    menu.setMnemonic(KeyEvent.VK_G);
    menu.getAccessibleContext().setAccessibleDescription("Start games");
    menuBar.add(menu);

    menuItem = new JMenuItem("Start Local Game", KeyEvent.VK_L);
    menuItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(final ActionEvent e)
      {
        treatStartLocalGameMenuItemClicked(e);
      }
    });
    menu.add(menuItem);
    menuItem = new JMenuItem("Start Networked Game", KeyEvent.VK_N);
    menuItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(final ActionEvent e)
      {
        treatStartNetworkedGameMenuItemClicked(e);
      }
    });
    menu.add(menuItem);

    menu = new JMenu("Help");
    menu.setMnemonic(KeyEvent.VK_H);
    menu.getAccessibleContext().setAccessibleDescription("Help");
    menuItem = new JMenuItem("Help Topics", KeyEvent.VK_T);
    menuItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(final ActionEvent e)
      {
        treatHelpTopicsMenuItemClicked(e);
      }
    });

    menu.add(menuItem);
    menu.addSeparator();
    menuItem = new JMenuItem("About", KeyEvent.VK_A);
    menuItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(final ActionEvent e)
      {
        treatAboutMenuItemClicked(e);
      }
    });
    menu.add(menuItem);

    menuBar.add(menu);

    _frame.setJMenuBar(menuBar);

    panel.add(_boardPanel, BorderLayout.CENTER);
    panel.add(_infoLabel, BorderLayout.SOUTH);
    _frame.getContentPane().add(panel);
    /// _frame.setContentPane(_boardPanel);
    _frame.pack();
    _frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
    _frame.addWindowListener(new WindowAdapter()
    {
      /**
       * Allow the user to confirm wether he wants to close or not.
       * @param e
       */
      public void windowClosing(WindowEvent e)
      {
        _logger.debug("windowClosing()");
        boolean shouldClose = true;
        if (_askCloseConfirm)
        {
          String confirmQuitMsg = "";
          if (_game != null)
          {
            confirmQuitMsg += "A game is being played. \n";
          }
          confirmQuitMsg += "Are you sure you want to quit?";
          int userChoice =
              JOptionPane.showConfirmDialog(JSetClientUI.this, confirmQuitMsg,
                                            "Confirm close", JOptionPane.YES_NO_OPTION);
          shouldClose = (userChoice == JOptionPane.OK_OPTION);
        }
        if (shouldClose)
        {
          onWindowClose();
        }
      }
    });
    _frame.setSize(800, 600);
    _frame.setBounds(50, 50, 800, 600);

    _frame.show();
  }

  private void onWindowClose()
  {
    if (_game != null)
    {
      _logger.debug("closing game...");
      _game.stop();
    }
    _logger.debug("Closing.");
    setVisible(false);
    dispose();
    System.exit(0);
  }

  private void setGame(final org.cb.jset.client.SetGame game)
  {
    if (_game == null)
    {
      _game = game;
      _game.addBoardListener(_gameController);
      _game.addGameListener(_gameController);
      _game.start();
    }
  }

  private void addCards(final CardProperties[] cardProperties)
  {
    _boardModel.addCards(cardProperties);
    updateInfoLabel();
    this.validate();
  }

  void updateInfoLabel()
  {
    final MatchingSetFinder lFinder = new MatchingSetFinder();
    final CardSet[] lFoundSets = lFinder.findSets(_boardModel.getElements());
    for (int i = 0; i < lFoundSets.length; i++)
    {
      final CardSet lFoundSet = lFoundSets[i];
      _logger.debug("foundSet[" + i + "]" + lFoundSet);
    }
    _infoLabel.setText("There are " + lFoundSets.length + " set(s) currently on the board");
  }

  /**
   * An object to display empty card space on the board.
   */
  private static class EmptyCard extends JPanel
  {
    public EmptyCard()
    {
      init();
    }

    void init()
    {
      // FIXME coded shared with JSetCardComponent. parametrize.
      final int width = 80;
      final int height = (int) (1.5 * width);
      setSize(width, height);
      setPreferredSize(getSize());
    }
  }


  /**
   * Control events coming from the board.
   */
  class BoardController implements CardBoardModelListener, SetGameClientCardListener, Serializable
  {
    public void intervalAdded(final CardBoardEvent e)
    {
      _logger.debug(e);
      final CardBoardEvent.IndexInterval interval = (CardBoardEvent.IndexInterval) e.getBoardIndexes();
      for (int i = interval.getIndex0(); i <= interval.getIndex1(); i++)
      {
        final JComponent cardComponent = getCardComponent(i);
        _boardPanel.add(cardComponent, i);
      }
      _boardPanel.validate();
    }

    public void intervalRemoved(final CardBoardEvent e)
    {
      _logger.debug(e);
      final CardBoardEvent.IndexInterval interval = (CardBoardEvent.IndexInterval) e.getBoardIndexes();
      for (int i = interval.getIndex0(); i <= interval.getIndex1(); i++)
      {
        final JPanel jPanel = new EmptyCard();
        replaceCardComponent(i, jPanel);
      }
      _boardPanel.validate();
    }

    public void contentsChanged(final CardBoardEvent e)
    {
      _logger.debug(e);
      final CardBoardEvent.IndexList indexList = (CardBoardEvent.IndexList) e.getBoardIndexes();
      for (int i = 0; i < indexList.getIndexesInBoard().size(); i++)
      {
        final Integer index = (Integer) indexList.getIndexesInBoard().get(i);
        final JSetCard card = (JSetCard) _boardModel.getElementAt(index.intValue());
        final JComponent newCard;
        if (card == null)
        {
          newCard = new EmptyCard();
        }
        else
        {
          newCard = getCardComponent(index.intValue());
        }
        replaceCardComponent(index.intValue(), newCard);
      }
      _boardPanel.validate();
    }

    public void notifyCardEvent(final CardEvent e)
    {
      _logger.debug(e);
      final int index = _boardModel.getElements().indexOf(e.getSource());
      final JSetCardComponent cardComponent = getCardComponent(index);
      cardComponent.updateBorder();
      if (cardComponent.isSelected())
      {
        _selectedCards.add(cardComponent.getCard());
        _logger.debug("selecting " + cardComponent.getCard() + " still " + _selectedCards.size());
      }
      else
      {
        _selectedCards.remove(cardComponent.getCard());
        _logger.debug("deselecting " + cardComponent.getCard() + " still " + _selectedCards.size());
      }

      if (_selectedCards.size() == 3)
      {
        final CardProperties[] cardProperties = new CardProperties[3];
        int i = 0;
        for (Iterator iterator = _selectedCards.iterator(); iterator.hasNext();)
        {
          final org.cb.jset.JSetCard card = (JSetCard) iterator.next();
          cardProperties[i++] = card.getProperties();
        }
        tryToMatch(new CardSet(cardProperties));
      }
      else if (_selectedCards.size() > 3)
      {
        throw new IllegalStateException("More than 3 cards selected");
      }
    }
  }

  /**
   * Control events coming from the game instance.
   *
   * FIXME we probably need to add multithreading handling on the model.
   */
  class GameController implements SetGameBoardListener, SetGameListener, Serializable
  {
    /**
     * @inheritDoc
     */
    public void cardsAdded(final CardProperties[] cards)
    {
      addCards(cards);
    }

    /**
     * @inheritDoc
     */
    public void setRemoved(final CardSet set)
    {
      treatSuccessfulMatching(set);
    }

    /**
     * @inheritDoc
     */
    public void gameChange(final GameEvent event)
    {
      // FIXME
    }
  }

  private void replaceCardComponent(final int i, final JComponent jPanel)
  {
    _boardPanel.remove(i);
    _boardPanel.add(jPanel, i);
  }

  private JSetCardComponent getCardComponent(final int i)
  {
    final org.cb.jset.JSetCard card = (JSetCard) _boardModel.getElementAt(i);
    JSetCardComponent cardComponent = (JSetCardComponent) _mapCardsToComponents.get(card);
    // FIXME was that really necessary ?
    if (cardComponent == null)
    {
      cardComponent = new JSetCardComponent(card);
      cardComponent.addCardEventListener(_boardController);
      _mapCardsToComponents.put(card, cardComponent);
    }
    // FIXME empty cache
    return cardComponent;
  }

  private void tryToMatch(final CardSet set)
  {
    boolean isMatchingSet;
    try
    {
      _game.removeSet(set);
      isMatchingSet = true;
    }
    catch (MatchingException e)
    {
      isMatchingSet = false;
    }

    if (!isMatchingSet)
    {
      treatFailedMatching(set, set.getCards());
    }
  }

  private void treatFailedMatching(final CardSet set, final CardProperties[] cardsToHandle)
  {
    _logger.debug("not matching set!!" + set);
    for (int j = 0; j < cardsToHandle.length; j++)
    {
      final JSetCardComponent jSetCardComponent = (JSetCardComponent) _mapCardsToComponents.get(new JSetCard(cardsToHandle[j]));
      jSetCardComponent.setSelection(false);
    }
  }

  private void treatSuccessfulMatching(final CardSet set)
  {
    _logger.debug("matching set!! " + set);
    // note: we don't replace, but remove then add new ones.
    try
    {
      _boardModel.removeSet(set);
    }
    catch (BoardException e)
    {
      throw new IllegalStateException("We shouldn't fail here... " + e.getMessage());
    }
    updateInfoLabel();
    // FIXME this shouldn't be there, but is there a clean way to clean cards we've lost track of?
    _selectedCards.clear();
//    if (! _cardStack.empty())
//    {
//      CardProperties[] lCardProperties = _cardStack.pop(3);
//      _boardModel.addCards(lCardProperties);
//    }
  }

  private void treatAboutMenuItemClicked(final ActionEvent e)
  {
    final String aboutMsg = "Implementation of the JSet Card Board game.\n"
     + "Copyright (C) Jerome Lacoste (jerome@coffeebreaks.org) 2004.\n"
     + "V. " + JSetBuild.VERSION + " built " + JSetBuild.BUILD_TIME;
    JOptionPane.showMessageDialog(JSetClientUI.this, aboutMsg);
  }
  private void treatHelpTopicsMenuItemClicked(final ActionEvent e)
  {

  }
  private void treatStartLocalGameMenuItemClicked(final ActionEvent e)
  {
    setGame(new LocalGame());
  }

  private void treatStartNetworkedGameMenuItemClicked(final ActionEvent e)
  {
    if (System.getSecurityManager() == null)
    {
      System.setSecurityManager(new RMISecurityManager());
    }
    try
    {
      final String host = "localhost";
      final String gameName = "TestGame";
      final RemoteSetGame game = Client.getGame(host, gameName);
      final NetworkGame networkGame = new NetworkGame("UITestUser");
      networkGame.setRemoteGame(game);
      setGame(networkGame);
    }
    catch (Exception e1)
    {
      new ErrorHandler().handle(e1);
    }
  }

  class ErrorHandler implements Serializable
  {
    void handle(final Exception e)
    {
      final String msg;
      if (e instanceof RemoteException)
      {
        msg = "Error while trying to connect :" + e.getMessage();
      }
      else if (e instanceof NotBoundException)
      {
        msg = "Error while trying to connect :" + e.getMessage();
      }
      else if (e instanceof MalformedURLException)
      {
        msg = "Error while trying to connect :" + e.getMessage();
      }
      else
      {
        msg = "Unkown error :" + e.getMessage();
      }
      JOptionPane.showMessageDialog(JSetClientUI.this, msg);
    }
  }

  public static void main(final String[] args)
  {
    org.apache.log4j.PropertyConfigurator.configure("src/conf/log4j.properties");
    final JSetClientUI lDrawing = new JSetClientUI("JSet - " + JSetBuild.VERSION);
  }
}
