// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset.server;

import org.apache.log4j.Logger;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;
import java.util.List;
import java.net.MalformedURLException;

/**
 * Provides accesses to RemoteSetGame implementations.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: JSetGameServer.java 130 2004-04-15 05:18:07Z jerome $
 */
public class JSetGameServer extends UnicastRemoteObject implements RemoteSetGameServer
{
  private static final Logger LOGGER = Logger.getLogger(JSetGameServer.class);

  /**
   * A class used to register the different creates games.
   */
  private static class Games
  {
    private final List _games = new Vector();

    private synchronized void addGame(final RemoteSetGame game)
    {
      _games.add(game);
    }

    /**
     * @param game
     * @todo ending remote games should deregister themselves
     */
    private synchronized void removeGame(final RemoteSetGame game)
    {
      _games.remove(game);
    }

    private synchronized RemoteSetGame[] getGames()
    {
      RemoteSetGame[] games = new RemoteSetGame[_games.size()];
      games = (RemoteSetGame[]) _games.toArray(games);
      return games;
    }

    private synchronized RemoteSetGame getGame(final int id) throws RemoteException
    {
      for (int i = 0; i < _games.size(); i++)
      {
        final RemoteSetGame game = (RemoteSetGame) _games.get(i);
        if (game.getID() == id)
        {
          return game;
        }
      }
      return null;
    }

    private void closeGames()
    {
      for (int i = 0; i < _games.size(); i++)
      {
        final JSetRemoteGameImpl game = (JSetRemoteGameImpl) _games.get(i);
        game.terminate();
      }
    }
  }

  private final Games _games = new Games();

  /**
   * Creates a new instance of JSetGameServer.
   * @throws RemoteException if failed to export object
   * @todo specify a port
   */
  public JSetGameServer() throws RemoteException
  {
    addShutdownHook();
  }

  /**
   * Create a game with the specified number of players.
   * @param name the name of the game.
   * @param nbPlayers the number of players.
   * @return the new game id
   * @throws java.rmi.RemoteException upon communication related exception on this remote call
   */
  public int createGame(final String name, final int nbPlayers) throws java.rmi.RemoteException
  {
    // FIXME search for games with existing name
    final RemoteSetGame game = new JSetRemoteGameImpl(name, nbPlayers);
    _games.addGame(game);
    return game.getID();
  }

  /**
   * Returns the game identified by the specified id.
   * @param id the id of the game to return
   * @return the game if found, or <code>null</code> if none found.
   * @throws java.rmi.RemoteException upon communication related exception on this remote call
   */
  public RemoteSetGame getGame(final int id) throws RemoteException
  {
    return _games.getGame(id);
  }

  /**
   * @return the different remote games known to this server.
   * @throws java.rmi.RemoteException upon communication related exception on this remote call
   */
  public RemoteSetGame[] getGames() throws RemoteException
  {
    return _games.getGames();
  }

  /**
   * Make sure all games are closed when this game server dies.
   */
  private void addShutdownHook()
  {
    /**
     * a Runnable that close all games.
     */
    class JSetShutdownHook implements Runnable
    {
      /**
       * run this runnable
       */
      public void run()
      {
        LOGGER.debug("ShutdownHook: dying.");
        _games.closeGames();
      }
    }
    LOGGER.debug("Registering shutdown hook");
    Runtime.getRuntime().addShutdownHook(new Thread(new JSetShutdownHook()));
  }

  /**
   * Starts a JSetGameServer.
   * @param args no specific argument expected
   */
  public static void main(final String[] args)
  {
    //java.util.Properties props = System.getProperties();
    //props.list(System.out);
    org.apache.log4j.PropertyConfigurator.configure("log4j.properties");
    LOGGER.info("CLASSPATH:" + System.getProperty("java.class.path"));
    if (System.getSecurityManager() == null)
    {
      LOGGER.debug("adding RMISecurityManager");
      System.setSecurityManager(new java.rmi.RMISecurityManager());
    }
    final String name = "//localhost/JSetServer";
    RemoteSetGameServer server;
    try
    {
      server = new JSetGameServer();
    }
    catch (RemoteException e)
    {
      LOGGER.error("JSetGameServer exception: failure to export game server object.", e);
      System.exit(-1);
      return; // please compiler
    }
    try
    {
      java.rmi.Naming.rebind(name, server);
      LOGGER.info("JSetGameServer bound");
    }
    catch (RemoteException e)
    {
      LOGGER.error("JSetGameServer exception: couldn't connect to registry."
      + " Check that the rmiregistry is running on the specified host.", e);
      System.exit(-1);
    }
    catch (MalformedURLException e)
    {
      LOGGER.error("JSetGameServer exception: check your config " + name, e);
      System.exit(-1);
    }
  }
}
