// START LICENSE
// JSet - a Java JSet card board game implementation
// Copyright (C) 2004 Jerome Lacoste
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// END LICENSE

package org.cb.jset.server;

import org.apache.log4j.Logger;
import org.cb.jset.SetGameConnection;
import org.cb.jset.JSetGameImpl;
import org.cb.jset.CardSet;
import org.cb.jset.MatchingException;
import org.cb.jset.SetGamePlayer;
import org.cb.jset.CardProperties;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.Serializable;

/**
 * An implementation of a remote Game.
 * <p>
 * Contains several inner classes that act as wrappers, allowing to transparently use remote implementations locally.
 *
 * @author jerome@coffeebreaks.org - last modified by $LastChangedBy: jerome $
 * @version $Id: JSetRemoteGameImpl.java 130 2004-04-15 05:18:07Z jerome $
 */
public class JSetRemoteGameImpl extends UnicastRemoteObject implements RemoteSetGame
{
  private final Logger _logger = Logger.getLogger(this.getClass().getName());

  private final JSetGameImpl _game;

  /**
   * Constructs a remote game instance given the specified name and number of players.
   * @param name
   * @param nbPlayers
   * @throws java.rmi.RemoteException upon communication related exception on this remote call
   */
  JSetRemoteGameImpl(final String name, final int nbPlayers) throws RemoteException
  {
    super();
    _game = new JSetGameImpl(name, nbPlayers);
    init();
  }

  /**
   * A remote Game connection that effectively wrapps a local one.
   */
  static class LocalConnectionWrapper extends UnicastRemoteObject implements Serializable, RemoteSetGameConnection
  {
    private final transient SetGameConnection _localConnection;

    /**
     * Constructs an instance of <code>LocalConnectionWrapper</code>
     * @param localConnection
     * @throws java.rmi.RemoteException upon communication related exception on this remote call
     */
    public LocalConnectionWrapper(final SetGameConnection localConnection) throws RemoteException
    {
      _localConnection = localConnection;
    }

    /**
     * Used by a registered player to try to match a set of set.
     * @param set the set to try to match.
     * @throws MatchingException if the set doesn't match or if a card is not on the board.
     * @see org.cb.jset.SetGameConnection#matchSet(org.cb.jset.CardSet)
     */
    public void matchSet(final CardSet set) throws MatchingException
    {
      _localConnection.matchSet(set);
    }

    /**
     * Close the game connection, effectively leaving the game.
     * @see org.cb.jset.SetGameConnection#close()
     */
    public void close()
    {
      _localConnection.close();
    }

    /**
     * Say wether the game connection is closed or not.
     * @return <code>true</code> if the connection is closed, <code>false</code> otherwise.
     * @see org.cb.jset.SetGameConnection#isClosed()
     */
    public boolean isClosed()
    {
      return _localConnection.isClosed();
    }
  }

  /**
   * Implementation of a player for a remote game.
   * Effectively wraps a {@link RemoteSetGamePlayer}.
   */
  static class RemotePlayerWrapper implements SetGamePlayer
  {
    private final RemoteSetGamePlayer _remotePlayer;
//    private boolean _wasDisconnected;
//    private RemoteException _remoteException;

    /**
     * Constructs an instance using the specified remotePlayer.
     * @param remotePlayer
     */
    public RemotePlayerWrapper(final RemoteSetGamePlayer remotePlayer)
    {
      _remotePlayer = remotePlayer;
    }

    private void treatConnectionProblem(final RemoteException e)
    {
//      _wasDisconnected = true;
//      _remoteException = e;
      throw new IllegalStateException("Failure to make a remote call... " + e.getMessage());
    }

    /**
     * @return the name of the player.
     */
    public String getName()
    {
      try
      {
        return _remotePlayer.getName();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return ""; // please compiler...
    }

    /**
     * Called by a game when the player should be blocked.
     * @see #isBlocked()
     */
    public void block()
    {
      try
      {
        _remotePlayer.block();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
    }

    /**
     * Called by a game when the player should be unblocked.
     * @see #isBlocked()
     */
    public void unblock()
    {
      try
      {
        _remotePlayer.unblock();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
    }

    /**
     * A blocked player doesn't have the right to play.
     * @return <code>true</code> is the player is blocked, <code>false</code> otherwise.
     */
    public boolean isBlocked()
    {
      try
      {
        return _remotePlayer.isBlocked();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
      return false; // please compiler...
    }


    /**
     * Called to notify a disconnection from a game.
     */
    public void disconnect()
    {
      try
      {
        _remotePlayer.disconnect();
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
    }

    /**
     * Called to notify that the specified cards were added to the board.
     * @param cards the cards that were added to the board.
     */
    public void cardsAdded(final CardProperties[] cards)
    {
      try
      {
        _remotePlayer.cardsAdded(cards);
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
    }

    /**
     * Called to notify that the specified set was removed from the board.
     * The set should be a {@link CardSet#isMatching() matching set}.
     * @param set the set that was removed from the board
     */
    public void setRemoved(final CardSet set)
    {
      try
      {
        _remotePlayer.setRemoved(set);
      }
      catch (RemoteException e)
      {
        treatConnectionProblem(e);
      }
    }
  }

  void init()
  {
    _game.start();
  }

  void terminate()
  {
    _game.terminate();
  }

  /**
   * @inheritDoc
   */
  public String getName() throws RemoteException
  {
    return _game.getName();
  }

  /**
   * @inheritDoc
   */
  public int getCurrentNbPlayers() throws RemoteException
  {
    return _game.getCurrentNbPlayers();
  }

  /**
   * @inheritDoc
   */
  public int getMaxNbPlayers() throws RemoteException
  {
    return _game.getMaxNbPlayers();
  }

  /**
   * @inheritDoc
   */
  public RemoteSetGameConnection connect(final RemoteSetGamePlayer player) throws RemoteException
  {
    if (_logger.isInfoEnabled())
    {
      _logger.info("Connecting player:" + player.getName() + ">");
    }
    final SetGameConnection conn;
    synchronized (this)
    {
      final SetGamePlayer playerWrapper = new RemotePlayerWrapper(player);
      conn = _game.connect(playerWrapper);
    }
    return new LocalConnectionWrapper(conn);
  }

  /**
   * @inheritDoc
   */
  public int getID() throws RemoteException
  {
    return _game.getID();
  }
}
